@@ -1,4 +1,4 @@
-/*	$Id: context.c,v 1.12 2003/12/11 11:55:06 steve Exp $	*/
+/*	$Id: context.c,v 1.14 2004/09/12 19:20:28 steve Exp $	*/
 
 /*-
  * Copyright (c) 2001 Steve C. Woodford.
@@ -140,8 +140,12 @@ context_add_client(struct context *c, struct client_ctx *cc)
 			return (-1);
 
 		(void) snprintf(buff,
-		    sizeof(buff), "Connection to service `%s' from `%s'\r\n",
-		    c->c_name, client_getname(cc));
+		    sizeof(buff), "Connected to service `%s'\r\n", c->c_name);
+		if (client_output(cc, buff, strlen(buff), NULL) < 0)
+			return (-1);
+
+		(void) snprintf(buff,
+		    sizeof(buff), "From `%s'\r\n", client_getname(cc));
 		if (client_output(cc, buff, strlen(buff), NULL) < 0)
 			return (-1);
 
@@ -241,7 +245,8 @@ context_setup_pollfds(struct context *c, struct pollfd *pf)
 
 	if (c->c_server != NULL &&
 	    (events = ((client_read_pending(c->c_server) ? POLLRDNORM : 0) |
-	    (client_write_pending(c->c_server) ? POLLWRNORM : 0))) != 0) {
+	    (client_write_pending(c->c_server) ? POLLWRNORM : 0) |
+	    (c->c_server->cc_canhup ? POLLHUP : 0))) != 0) {
 		pf->events = events;
 		pf->revents = 0;
 		pf->fd = client_getfd(c->c_server);
@@ -253,7 +258,8 @@ context_setup_pollfds(struct context *c, struct pollfd *pf)
 	for (cc = TAILQ_FIRST(&c->c_clients); cc != NULL;
 	    cc = TAILQ_NEXT(cc, cc_qent)) {
 		if ((events = ((client_read_pending(cc) ? POLLRDNORM : 0) |
-		    (client_write_pending(cc) ? POLLWRNORM : 0))) == 0)
+		    (client_write_pending(cc) ? POLLWRNORM : 0) |
+		    (cc->cc_canhup ? POLLHUP : 0))) == 0)
 			continue;
 		pf->events = events;
 		pf->revents = 0;
