@@ -1,4 +1,4 @@
-/*	$Id: tty.c,v 1.12 2003/12/11 11:55:06 steve Exp $	*/
+/*	$Id: tty.c,v 1.14 2004/09/12 19:31:08 steve Exp $	*/
 
 /*-
  * Copyright (c) 2001 Steve C. Woodford.
@@ -32,14 +32,16 @@
 #include <sys/types.h>
 #include <sys/filio.h>
 #include <sys/ioctl.h>
-#include <sys/types.h>
 
+#include <assert.h>
 #include <ctype.h>
 #include <err.h>
 #include <errno.h>
 #include <fcntl.h>
+#include <paths.h>
 #include <stdlib.h>
 #include <string.h>
+#include <syslog.h>
 #include <termios.h>
 #include <unistd.h>
 
@@ -83,6 +85,7 @@ struct tty_opt {
 struct tty_ctx {
 	struct tty_opt		tc_to;
 	struct buffer_ctx	*tc_buff;
+	int			tc_open;
 };
 
 static struct tty_opt defopt = {
@@ -183,8 +186,16 @@ tty_init(struct client_ctx *cc, const void *arg)
 
 	cc->cc_data = tc;
 	tc->tc_to = *((const struct tty_opt *) arg);
+	cc->cc_fd = open(_PATH_DEVNULL, O_RDWR);
+	if (cc->cc_fd < 0) {
+		(void) free(tc->tc_to.to_device);
+		(void) free(tc);
+		return (-1);
+	}
 
-	if (tty_open_device(cc) < 0) {
+	if (tty_open_device(cc) < 0 ||
+	    buffer_init(&tc->tc_buff, cc->cc_fd) < 0) {
+		(void) close(cc->cc_fd);
 		(void) free(tc->tc_to.to_device);
 		(void) free(tc);
 		return (-1);
@@ -199,6 +210,7 @@ tty_destroy(struct client_ctx *cc)
 	struct tty_ctx *tc = cc->cc_data;
 
 	tty_close_device(cc);
+	(void) close(cc->cc_fd);
 	(void) free(tc->tc_to.to_device);
 	(void) free(tc);
 }
@@ -211,7 +223,7 @@ tty_event(struct client_ctx *cc, int events)
 	ssize_t rv;
 	int nbytes;
 
-	if (tc->tc_buff == NULL)
+	if (tc->tc_open == 0)
 		return (0);
 
 	if (events & POLLRDNORM) {
@@ -247,7 +259,7 @@ tty_output(struct client_ctx *cc, const char *buff, size_t len,
 {
 	struct tty_ctx *tc = cc->cc_data;
 
-	if (sender->cc_options.co_readonly || tc->tc_buff == NULL)
+	if (sender->cc_options.co_readonly || tc->tc_open == 0)
 		return (0);
 
 	return buffer_fill(tc->tc_buff, buff, len);
@@ -258,10 +270,7 @@ tty_read_pending(struct client_ctx *cc)
 {
 	struct tty_ctx *tc = cc->cc_data;
 
-	if (tc->tc_buff == NULL)
-		return (0);
-
-	return (1);
+	return (tc->tc_open);
 }
 
 static int
@@ -269,7 +278,7 @@ tty_write_pending(struct client_ctx *cc)
 {
 	struct tty_ctx *tc = cc->cc_data;
 
-	if (tc->tc_buff == NULL)
+	if (tc->tc_open == 0)
 		return (0);
 
 	return (buffer_length(tc->tc_buff) != 0);
@@ -283,14 +292,14 @@ tty_ioctl(struct client_ctx *cc, int cmd, void *arg, struct client_ctx *sender)
 
 	switch (cmd) {
 	case TTY_IOCTL_SENDBREAK:
-		if (tc->tc_buff != NULL &&
+		if (tc->tc_open &&
 		    (sender->cc_options.co_readonly == 0 ||
 		     sender->cc_options.co_allowbreak))
 			tcsendbreak(cc->cc_fd, 0);
 		break;
 
 	case CONTEXT_IOCTL_ADD_CLIENT:
-		if (tc->tc_buff == NULL && tty_open_device(cc) < 0)
+		if (tc->tc_open == 0 && tty_open_device(cc) < 0)
 			return (-1);
 		break;
 
@@ -310,16 +319,21 @@ tty_open_device(struct client_ctx *cc)
 	struct tty_ctx *tc = cc->cc_data;
 	struct termios tent;
 	int flags;
+	int fd, fd2;
 
-	if (tc->tc_buff != NULL)
+	if (tc->tc_open)
 		return (0);
 
-	if ((cc->cc_fd = open(tc->tc_to.to_device,
-	    O_RDWR | O_NONBLOCK | O_EXCL)) < 0)
+	if ((fd = open(tc->tc_to.to_device, O_RDWR | O_NONBLOCK | O_EXCL)) < 0){
+		syslog(LOG_ALERT, "tty_open_device(): failed to open %s (%d)",
+		    tc->tc_to.to_device, errno);
 		return (-1);
+	}
 
-	if (tcgetattr(cc->cc_fd, &tent) < 0) {
-		(void) close(cc->cc_fd);
+	if (tcgetattr(fd, &tent) < 0) {
+		syslog(LOG_ALERT, "tty_open_device(): tcgetattr() failed (%d)",
+		    errno);
+		(void) close(fd);
 		return (-1);
 	}
 
@@ -327,7 +341,9 @@ tty_open_device(struct client_ctx *cc)
 
 	if (cfsetispeed(&tent, tc->tc_to.to_ispeed) < 0 ||
 	    cfsetospeed(&tent, tc->tc_to.to_ospeed) < 0) {
-		(void) close(cc->cc_fd);
+		syslog(LOG_ALERT, "tty_open_device(): setspeed failed (%d)",
+		    errno);
+		(void) close(fd);
 		return (-1);
 	}
 
@@ -342,8 +358,10 @@ tty_open_device(struct client_ctx *cc)
 	tent.c_cflag &= tc->tc_to.to_cflagmask;
 	tent.c_cflag |= tc->tc_to.to_cflag;
 
-	if (tcsetattr(cc->cc_fd, TCSANOW, &tent) < 0) {
-		(void) close(cc->cc_fd);
+	if (tcsetattr(fd, TCSANOW, &tent) < 0) {
+		syslog(LOG_ALERT, "tty_open_device(): tcsetattr() failed (%d)",
+		    errno);
+		(void) close(fd);
 		return (-1);
 	}
 
@@ -351,17 +369,25 @@ tty_open_device(struct client_ctx *cc)
 	 * If this is *not* a pseudo tty, ensure DTR is asserted.
 	 * Note: This relies on TIOCGFLAGS returning ENOTTY for pty(4)'s.
 	 */
-	if (ioctl(cc->cc_fd, TIOCGFLAGS, &flags) == 0 && errno == ENOTTY &&
-	    ioctl(cc->cc_fd, TIOCSDTR, 0) < 0) {
-		(void) close(cc->cc_fd);
+	if (ioctl(fd, TIOCGFLAGS, &flags) == 0 && errno == ENOTTY &&
+	    ioctl(fd, TIOCSDTR, 0) < 0) {
+		syslog(LOG_ALERT, "tty_open_device(): assert DTR failed (%d)",
+		    errno);
+		(void) close(fd);
 		return (-1);
 	}
 
-	if (buffer_init(&tc->tc_buff, cc->cc_fd) < 0) {
-		(void) close(cc->cc_fd);
+	fd2 = dup2(fd, cc->cc_fd);
+	(void) close(fd);
+
+	if (fd2 != cc->cc_fd) {
+		syslog(LOG_ALERT, "tty_open_device(): dup2() failed (%d)",
+		    errno);
 		return (-1);
 	}
 
+	tc->tc_open = 1;
+
 	return (0);
 }
 
@@ -369,14 +395,27 @@ static void
 tty_close_device(struct client_ctx *cc)
 {
 	struct tty_ctx *tc = cc->cc_data;
+	int fd, fd2;
 
-	if (tc->tc_buff == NULL)
+	if (tc->tc_open == 0)
 		return;
 
-	buffer_destroy(tc->tc_buff);
-	(void) close(cc->cc_fd);
+	if ((fd = open(_PATH_DEVNULL, O_RDWR)) < 0) {
+		syslog(LOG_ALERT, "tty_close_device(): failed to open %s (%d)",
+		    _PATH_DEVNULL, errno);
+		return;
+	}
 
-	tc->tc_buff = NULL;
+	fd2 = dup2(fd, cc->cc_fd);
+	close(fd);
+
+	if (fd2 != cc->cc_fd) {
+		syslog(LOG_ALERT, "tty_close_device(): dup2 failed (%d)",
+		    errno);
+		return;
+	}
+
+	tc->tc_open = 0;
 }
 
 /* ARGSUSED */
